local RW = LibStub("AceAddon-3.0"):GetAddon("Raid Watch")
local L = LibStub("AceLocale-3.0"):GetLocale("RW-2.0")

local AL = LibStub("AceLocale-3.0")
local LSM = LibStub("LibSharedMedia-3.0")

local Boss = {}
RW.bossPrototype = Boss
RW.bosses:SetDefaultModulePrototype(Boss)

-- Party module, have all the info about the current party
local Party
local Comm

function Boss:OnInitialize()
	self.defaultOptions = {enabled = true}
	self.pName = UnitName("player")
	self.pClass = select(2, UnitClass("player"))
	RW:RegisterBoss(self) -- Calls OnRegister for the boss module
	
	Party = RW:GetPlugin("Party")
	Comm = RW:GetPlugin("Comm")
end

function Boss:OnEnable()
	self:RegisterMessage("RW_BarFinished")
end

function Boss:OnDisable()
	-- Cleanup
	self.engageTime = nil
	if self.trackedSpells then wipe(self.trackedSpells) end
	self:UnscheduleAll()
	self:StopAllTimers()
	self:ResetAllIcons()
	self:SendMessage("RW_RemoveAllHealthFrames")
	self:SendMessage("RW_ShowRangeCheck", false)
end

function Boss:Start(trigger)
	RW:Msg(L.BOSS_ENGAGE:format(self:GetName()))
	self.engageTime = GetTime()
	self:Enable()
	if self.OnEngage then self:OnEngage(trigger) end
end

function Boss:End(wipe, trigger)
	if not wipe then
		RW:Msg(L.BOSS_WIN:format(self:GetName(), RW:TimeToStr(GetTime() - self.engageTime)))
		PlaySoundFile(LSM:Fetch("sound", RW.db.profile.vicotorySound))
	else
		RW:Msg(L.BOSS_WIPE:format(self:GetName(), RW:TimeToStr(GetTime() - self.engageTime)))
	end
	self:Disable()
	if self.OnEnd then self:OnEnd(wipe, trigger) end
end


function Boss:RegisterLocale(lang)
	return AL:NewLocale(self:GetName(), lang, true)
end

function Boss:GetLocale()
	return AL:GetLocale(self:GetName())
end

function Boss:AddSpellOption(category, ...)
	--[===[@debug@
	assert(select("#", ...) > 0, "AddSpellOption failed, no spells supplied")
	for i = 1, select("#", ...) do
		local s = select(i, ...)
		assert(type(s) == "number", ("Error registering spell option, %q is not a number"):format(s))
	end
	--@end-debug@]===]
	
	self.options = self.options or {}
	local name = RW.SpellName[select(1, ...)]
	if RW:InTable2(self.options, "text", name) then return end
	
	self:AddCategory(category)
	
	for i = 1, select("#", ...) do
		local spell = select(i, ...)
		local name = RW.SpellName[spell]
		if not RW:InTable2(self.options, "text", name) then
			self.defaultOptions[name] = {}
			self.defaultOptions[name].uses = 0 
			tinsert(self.options, {
				type = "spell",
				header = category,
				text = name,
				icon = select(3, GetSpellInfo(spell)),
				spell = spell,
				uses = 0
			})
		end
	end
end

function Boss:AddBoolOption(optionName, option, category, default)
	local options = self.options or {}
	if options[optionName] then return end
	
	self:AddCategory(category or "General")
	
	options[optionName] = {
		type = "bool",
		header = category or "General",
		text = option,
		optionName = optionName,
		enabled = (default == nil) or default
	}
	self.defaultOptions[optionName] = (default == nil) or default
	self.options = options
end

function Boss:AddSliderOption(optionName, option, category, default)
	local options = self.options or {}
	if options[optionName] then return end
	options[optionName] = {
		type = "slider",
		header = category,
		text = option,
		optionName = optionName,
	}
	self.defaultOptions[optionName] = default or 0
	self.options = options
end

function Boss:AddGeneralOption(option, category)
	local options = self.options or {}
	if options[option] then return end
	
	self:AddCategory(category or "General")
	
	options[option] = {
		type = "general",
		header = category or "General",
		text = option,
		uses = 0
	}
	self.defaultOptions[option] = {}
	self.defaultOptions[option].uses = 0
	self.options = options
end

function Boss:GetOpt(optName)
	return self.db.profile[optName]
end

function Boss:AddCategory(...)
	local options = self.options or {}
	options.headers = options.headers or {}
	for i = 1, select("#", ...) do
		local cat = select(i, ...)
		if not RW:InTable(options.headers, cat) then
			tinsert(options.headers, cat)
		end
	end
	self.options = options
end

function Boss:AddDefaults(optionName, oType, sType, default, customOptName)
	local defaultOption = self.defaultOptions[optionName]
	--local option = self.options[optionName]
	local option
	for i, o in pairs(self.options) do
		if o.text == optionName then 
			option = o 
			break
		end
	end
	-- Add the default state
	if ((default == nil) or default) then
		defaultOption.uses = RW:AddBit(defaultOption.uses, RW.BITS[oType])
		if sType ~= "CUSTOM" then
			defaultOption[oType] = RW:AddBit(defaultOption[oType], self.BITS[oType][sType])
		else
			defaultOption[oType..(customOptName or "Custom")] = ((default == nil) or default)
		end
	else
		defaultOption[oType] = RW:AddBit(defaultOption[oType], 0)
	end
	-- Add settings for so that UI can draw them no matter if they are default on or off
	if sType ~= "CUSTOM" then
		option[oType] = RW:AddBit(option[oType], self.BITS[oType][sType])
	else
		option["warning"..(customOptName or "Custom")] = ((default == nil) or default)
	end
	option.uses = RW:AddBit(option.uses, RW.BITS[oType])
end

function Boss:SetDelay(delay)
	self.endDelay = delay
end

function Boss:SetTrigger(...)
	for i = 1, select("#", ...) do
		local v = select(i, ...)
		if type(v) == "string" then
			self.yellTriggers = self.yellTriggers or {}
			self.yellTriggers[v] = true
		elseif type(v) == "number" then
			self.mobTriggers = self.mobTriggers or {}
			self.mobTriggers[v] = true
		end
	end
end

function Boss:SetWin(...)
	if not ... then
		self.mobWins = self.mobWins or {}
		self.mobWins[self:GetFirstTrigger()] = true
	elseif select(1, ...) == "combat" then
		self.winTrigger = "combat"
	end
	for i = 1, select("#", ...) do
		local v = select(i, ...)
		if type(v) == "string" then
			self.yellWins = self.yellWins or {}
			self.yellWins[v] = true
		elseif type(v) == "number" then
			self.mobWins = self.mobWins or {}
			self.mobWins[v] = true
		end
	end
end

function Boss:GetFirstTrigger()
	local cID
	for mob in pairs(self.mobTriggers) do
		cID = mob
		break
	end
	return cID
end

-------------------
-- General function
-------------------
function Boss:GetPartyType()
	return Party:GetPartyType()
end

function Boss:GetCUIDFromGUID(gUID)
	if not gUID then return end
	return tonumber(gUID:sub(9, 12), 16)
end

function Boss:GetUIDFromCID(cID)
	for name, member in pairs(Party.group) do
		if self:GetCUIDFromGUID(UnitGUID(member.id.."target")) == cID then
			return member.id.."target"
		end
	end
end

function Boss:GetCUIDFromUID(uID)
	if not uID then return end
	local gUID = UnitGUID(uID)
	if gUID then
		return self:GetCUIDFromGUID(gUID)
	end
end

function Boss:GetRaidTargetFromGUID(gUID)
	for name, member in pairs(Party.group) do
		if UnitGUID(member.id.."target") == gUID then
			return member.id.."target"
		end
	end
end

function Boss:GetTargetFromName(name)
	for name, member in pairs(Party.group) do
		if UnitName(member.id.."target") == name then
			return member.id.."target"
		end
	end
end

function Boss:UIDFromRandom(target)
	local targetID
	if target:find("^0x") then		-- Target by gUID
		targetID = self:GetRaidTargetFromGUID(target)
	elseif Party.group[target] then	-- Target by party member name
		targetID = Party.group[target].id
	else		
		targetID = self:GetTargetFromName(target)  -- Target by mob name
	end
	return targetID
end

function Boss:GetMobTarget(cID)
	if not cID then cID = self:GetFirstTrigger() end
	if not cID then return end
	if cID == Boss:GetCUIDFromUID("target") then return UnitName("targettarget") end
	for i, player in pairs(Party.group) do
		if cID == Boss:GetCUIDFromUID(player.id.."target") then return UnitName(player.id.."targettarget") end
	end
end

function Boss:GetBossHealth(cID)
	cID = cID or self:GetFirstTrigger()
	if not cID then return 0.0 end
	local uID = self:GetUIDFromCID(cID)
	if not uID then return 0.0 end
	return UnitHealth(uID) / UnitHealthMax(uID)
end

function Boss:IsHeroic()
	pType = self:GetPartyType()
	diff = GetInstanceDifficulty()
	if (pType == "party" and diff == 2) or diff > 2 then return true end
end

function Boss:InstanceSize()
	diff = GetInstanceDifficulty()
	if diff == 1 or diff == 3 then return 10
	else return 25 end
end

function Boss:SetIcon(icon, target)
	SetRaidTarget(target, icon or 0)
end

function Boss:Track(func, ...)
	if type("func") == "string" and not self[func] then
		error(("Missing function %q for bossmod %q"):format(func, self:GetName()))
	end
	self:SendMessage("RW_TrackRegistered", self, func, ...)
	self.trackedSpells = self.trackedSpells or {}
	for i = 1, select("#", ...) do
		self.trackedSpells[select(i, ...)] = func
	end
end

function Boss:YellTrack(func, ...)
	if type("func") == "string" and not self[func] then
		error(("Missing function %q for bossmod %q"):format(func, self:GetName()))
	end
	self:SendMessage("RW_YellRegistered", self, func, ...)
	self.trackedYells = self.trackedYells or {}
	for i = 1, select("#", ...) do
		self.trackedYells[select(i, ...)] = func
	end
end

-------------------

-------------
-- Scheduling
-------------
do
	local timers = {}
	local args = {}
		
	local function clearTimer(id)
		if not id or not timers[id] then return end
		local timer = timers[id]
		if timer.args then wipe(timer.args) end
		wipe(timer)
	end
	
	local function handleTimers(id)
		local timer = timers[id]
		local func = timer.func
		local id = timer.id
		local m = timer.module
		if timer.args then
			wipe(args)
			for i, v in pairs(timer.args) do tinsert(args, v) end
		end
		if type(timer.func) == "function" then
			func(unpack(args))
		else
			if m[func] then
				m[func](m, unpack(args))
			end
		end
		if not timer.rep then
			clearTimer(id)	
		end
	end
	
	local function prepareSchedule(self, id, time, func, ...)
		local timer = timers[id]
		if timer then
			self:Unschedule(id)
		else
			timer = {}
		end
		timer.func = func
		timer.module = self
		timer.args = {}
		if select("#", ...) > 0 then
			if timer.args then
				for i = 1, select("#", ...) do
					tinsert(timer.args, (select(i, ...)))
				end
			else
				timer.args = {...}
			end
		end
		timers[id] = timer
	end
	
	function Boss:Schedule(id, time, func, ...)
		prepareSchedule(self, id, time, func, ...)
		timers[id].id = self:ScheduleTimer(handleTimers, time, id)
	end
	
	function Boss:ScheduleRepeating(id, time, func, now, ...)
		prepareSchedule(self, id, time, func, ...)
		timers[id].rep = true
		timers[id].id = self:ScheduleRepeatingTimer(handleTimers, time, id)
		if now then handleTimers(id) end
	end

	function Boss:Unschedule(id)
		if timers[id] then
			self:CancelTimer(timers[id].id, true)
			clearTimer(id)
		end
	end

	function Boss:UnscheduleAll()
		for id, timer in pairs(timers) do
			if timer.module == self then
				self:Unschedule(id)
			end
		end
	end
end
--------------------------

-- Option helpers
Boss.optionHelpers = {}
function Boss.optionHelpers.onOptionEnter(self, event)
	local option, type, subType = self:GetUserData("option"), self:GetUserData("type"), self:GetUserData("subType")
	GameTooltip:SetOwner(self.frame)
	GameTooltip:AddLine(L[type.."_"..subType])
	GameTooltip:AddLine(L[type.."_"..subType.."_DESC"]:format(option.text))
	GameTooltip:Show()
end
function Boss.optionHelpers.onOptionLeave(self, event)
	GameTooltip:Hide()
end
function Boss.optionHelpers.onValueChanged(widget, event, ...)
	local value = ...
	local oType, oSubType, self, option = widget:GetUserData("type"), widget:GetUserData("subType"), widget:GetUserData("boss"), widget:GetUserData("option")
	local db = self.db.profile
	if not value then
		db[option.text][oType] = RW:RemoveBit(db[option.text][oType], self.BITS[oType][oSubType])
	else
		db[option.text][oType] = RW:AddBit(db[option.text][oType], self.BITS[oType][oSubType])
	end
end
function Boss.optionHelpers.onCustomValueChanged(widget, event, ...)
	local value = ...
	local boss, option, optName = widget:GetUserData("boss"), widget:GetUserData("option"), widget:GetUserData("optName")
	boss.db.profile[option.text][optName] = value
end
function Boss.optionHelpers.onValueChanged2(widget, event, ...)
	local key, subKey, self, option = widget:GetUserData("key"), widget:GetUserData("subKey"), widget:GetUserData("boss"), widget:GetUserData("option")
	local db = self.db.profile
	if select("#", ...) > 1 then
		db[option.text][key][subKey] = {...} -- Color
	else
		db[option.text][key][subKey] = ... -- True, False
	end
end
